/*
 * Startup code for DaVinci Initial Bootloader.
 *
 * Copyright (c) 2007 Sergey Kubushyn <ksi@koi8.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <config.h>

.arm
.globl	_start
.section	.boot
	.rept	64
	nop
	.endr

.org	0x100
_start:
	/*
	 * set the cpu to SVC mode
	 */
	mrs	r0, cpsr
	bic	r0, r0, $0x1f
	orr	r0, r0, $0xd3
	msr	cpsr, r0

	/* Set up the stack */
stack_setup:
	ldr	sp, c_stack		/* Fixed location for DaVinci */
	ldr	r0, c_stack_size
	add	sp, sp, r0

	/* Copy initialized data to where it belongs */
copy_data:
	ldr	r0, _text_end
	ldr	r1, _idata_start
	ldr	r2, _idata_end

cpdata_l:
	cmp	r2, r1
	beq	clear_bss
	ldr	r3, [r0]
	str	r3, [r1]
	add	r0, r0, $4
	add	r1, r1, $4
	b	cpdata_l

	/* Clear BSS section */
clear_bss:
	ldr	r0, _bss_start
	ldr	r1, _bss_end
	mov 	r2, $0x00000000

clbss_l:
	str	r2, [r0]
	add	r0, r0, $4
	cmp	r0, r1
	bls	clbss_l

	/*
	 * flush v4 I/D caches
	 */
	mov	r0, $0
	mcr	p15, 0, r0, c7, c7, 0
	mcr	p15, 0, r0, c8, c7, 0

	/*
	 * disable MMU and caches
	 */
	mrc	p15, 0, r0, c1, c0, 0
	bic	r0, r0, $0x00002300
	bic	r0, r0, $0x00000087
	orr	r0, r0, $0x00000002
	orr	r0, r0, $0x00001000
	mcr	p15, 0, r0, c1, c0, 0

	/*
	 * Go setup Memory and board specific bits.
	 */
	bl	platformsetup

	/* Branch to C main() */
	bl	main
loop4ever:
	b	loop4ever	

/* Our poor man's version of puts() */
.globl	printstr
printstr:
	ldr	r2, UART0_THR
printstr1:
	ldrb	r3, [r0], $1
	cmp	r3, $0
	moveq	pc, lr	
printstr2:
	ldr	r1, [r2, $0x14]
	tst	r1, $0x20 
	beq	printstr2
	str	r3, [r2]
	b	printstr1

/* Our poor man's version of putchar() */
.globl	putchr
putchr:
	ldr	r2, UART0_THR
putchr1:
	ldr	r1, [r2, $0x14]
	tst	r1, $0x20 
	beq	putchr1
	str	r0, [r2]
	mov	pc, lr

/* Our poor man's version of getchar() */
.globl	getchr
getchr:
	ldr	r1, UART0_THR
getchr1:
	ldr	r0, [r1, $0x14]
	tst	r0, $0x01 
	beq	getchr1
	ldr	r0, [r1]
	mov	pc, lr

/* Our poor man's version of gets() */
.globl	getstr
getstr:
	stmfd	sp!, {r4-r5, lr}	/* Enter */

	ldr	r4, str_buf
	mov	r5, $0
getstr_l:
	bl	getchr
	cmp	r5, $63
	beq	getstr_full	

	cmp	r0, $0x0d		/* CR (end of string, ENTER) */
	moveq	r0, $0
	streqb	r0, [r4, r5]
	beq	getstr_ret

	cmp	r0, $0x03		/* ^C, discard string, return -1 */
	moveq	r0, $0
	streqb	r0, [r4]
	moveq	r5, $-1
	beq	getstr_ret

	/* Only accept [0-9A-Za-z] and SPACE */
	cmp	r0, $' '	
	beq	getstr_get_next
	cmp	r0, $'/'
	bls	getstr_beep_skip
	cmp	r0, $'9'
	bls	getstr_get_next
	cmp	r0, $'@'
	bls	getstr_beep_skip
	cmp	r0, $'Z'
	bls	getstr_get_next
	cmp	r0, $'`'
	bls	getstr_beep_skip
	cmp	r0, $'z'
	bhi	getstr_beep_skip

	/* Everything else should be a normal character, fall through */	
getstr_get_next:
	bl	putchr
	strb	r0, [r4, r5]
	add	r5, r5, $1
	b	getstr_l

getstr_beep_skip:
	mov	r0, $0x07		/* Console beep */
	bl	putchr
	b	getstr_l

getstr_full:
	mov	r0, $0
	strb	r0, [r4, r5]
	mov	r0, $0x07		/* Console beep */
	bl	putchr
	mov	r5, $63

getstr_ret:
	mov	r0, $0x0d
	bl	putchr
	mov	r0, $0x0a
	bl	putchr
	mov	r0, r5
	ldmfd	sp!, {r4-r5, pc}	/* Leave */

/* Much easier to do it from assembly :) */
.globl	jump
jump:
	mov	pc, r0

.globl	load_raw
load_raw:
	stmfd	sp!, {r4-r5, lr}

	mov	r4, r0			/* Load address */
	movs	r5, r1			/* Size in bytes */
	beq	load_raw_end		/* Do nothing if zero size */

load_raw_loop:
	bl	getchr
	strb	r0, [r4], $1
	subs	r5, r5, $1
	bne	load_raw_loop

load_raw_end:
	ldmfd	sp!, {r4-r5, pc}

.globl	wait_loop
wait_loop:
	cmp	r0, $0
	moveq	pc, lr
	nop
	sub	r0, r0, $1
	b	wait_loop

c_stack:
	.word	0xbc00
c_stack_size:
	.word	0x400
.globl	_text_end
_text_end:
	.word	__text_end
.globl	_idata_start
_idata_start:
	.word	__idata_start
.globl	_idata_end
_idata_end:
	.word	__idata_end
_bss_start:
	.word	__bss_start
_bss_end:
	.word	__bss_end
UART0_THR:
	.word	0x01c20000
.globl	str_buf
str_buf:
	.word	__str_buf
.globl	_ldr_entry
_ldr_entry:
	.word	_start

.ltorg


.section	.bss
.globl	__str_buf
__str_buf:
	.rept	64
	.byte
	.endr
